﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System.Collections.Generic;

  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  /// <summary>
  ///     Duplicates the top item on the INTEGER stack. Does not pop its argument
  ///     (which, if it did, would negate the effect of the duplication!).
  /// </summary>
  /// <typeparam name="T">Stacktype</typeparam>
  [StorableClass]
  public abstract class DuplicateExpression<T> : StatelessExpression {
    protected DuplicateExpression() { }
    [StorableConstructor]
    protected DuplicateExpression(bool deserializing) : base(deserializing) { }

    protected void Eval(IPushStack<T> stack) {
      stack.Push(stack.Top);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.Integer, "INTEGER.DUP")]
  public class IntegerDuplicateExpression : DuplicateExpression<long> {
    public IntegerDuplicateExpression() { }
    [StorableConstructor]
    protected IntegerDuplicateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.IntegerStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.IntegerStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.Float, "FLOAT.DUP")]
  public class FloatDuplicateExpression : DuplicateExpression<double> {
    public FloatDuplicateExpression() { }
    [StorableConstructor]
    protected FloatDuplicateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.FloatStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.Boolean, "BOOLEAN.DUP")]
  public class BooleanDuplicateExpression : DuplicateExpression<bool> {
    public BooleanDuplicateExpression() { }
    [StorableConstructor]
    protected BooleanDuplicateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.BooleanStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.BooleanStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.Name, "NAME.DUP")]
  public class NameDuplicateExpression : DuplicateExpression<string> {
    public NameDuplicateExpression() { }
    [StorableConstructor]
    protected NameDuplicateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.NameStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.NameStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.Exec, "EXEC.DUP", execIn: 1)]
  public class ExecDuplicateExpression : DuplicateExpression<Expression> {
    public ExecDuplicateExpression() { }
    [StorableConstructor]
    protected ExecDuplicateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.ExecStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.ExecStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.Code, "CODE.DUP")]
  public class CodeDuplicateExpression : DuplicateExpression<Expression> {
    public CodeDuplicateExpression() { }
    [StorableConstructor]
    protected CodeDuplicateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.CodeStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.CodeStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.Char, "CHAR.DUP")]
  public class CharDuplicateExpression : DuplicateExpression<char> {
    public CharDuplicateExpression() { }
    [StorableConstructor]
    protected CharDuplicateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.CharStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.CharStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.String, "STRING.DUP")]
  public class StringDuplicateExpression : DuplicateExpression<string> {
    public StringDuplicateExpression() { }
    [StorableConstructor]
    protected StringDuplicateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.StringStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.IntegerVector, "INTEGER[].DUP")]
  public class IntegerVectorDuplicateExpression : DuplicateExpression<IReadOnlyList<long>> {
    public IntegerVectorDuplicateExpression() { }
    [StorableConstructor]
    protected IntegerVectorDuplicateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.IntegerVectorStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.IntegerVectorStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.FloatVector, "FLOAT[].DUP")]
  public class FloatVectorDuplicateExpression : DuplicateExpression<IReadOnlyList<double>> {
    public FloatVectorDuplicateExpression() { }
    [StorableConstructor]
    protected FloatVectorDuplicateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatVectorStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.FloatVectorStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.BooleanVector, "BOOLEAN[].DUP")]
  public class BooleanVectorDuplicateExpression : DuplicateExpression<IReadOnlyList<bool>> {
    public BooleanVectorDuplicateExpression() { }
    [StorableConstructor]
    protected BooleanVectorDuplicateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.BooleanVectorStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.BooleanVectorStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.StringVector, "STRING[].DUP")]
  public class StringVectorDuplicateExpression : DuplicateExpression<IReadOnlyList<string>> {
    public StringVectorDuplicateExpression() { }
    [StorableConstructor]
    protected StringVectorDuplicateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringVectorStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.StringVectorStack);
    }
  }
}